/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Receives a bunch of messages and identifies messages that are correlated (XPATH expression). 
 * <p>
 * Once a complete set of messages has been received (XPATH
 * expression on special message to decide that the set is complete), the
 * aggregator collects information from each correlated message and publishes a
 * single, aggregated message to the output channel for further processing. It
 * returns the result to this special message if it exists.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class Aggregator extends AbstractPattern {

    public final static String AGGREGATOR_NAMESPACE = EIPConstants.NAMESPACE + "/aggregator";

    private final Map<String, List<NormalizedMessage>> inMessagesMap = new ConcurrentHashMap<String, List<NormalizedMessage>>();

    /**
     * Build the aggregate Source from list of messages.
     * 
     * @throws JBIException
     */
    private static final Source buildContent(final List<NormalizedMessage> inMessagesList,
            final ExchangeContext context) throws MessagingException, PEtALSCDKException {
        Document aggregatedDocument = null;
        for (final NormalizedMessage inMessage : inMessagesList) {
            aggregatedDocument = PatternHelper.aggregateContent(inMessage, aggregatedDocument,
                    EIPConstants.RESULT_ELEMENT, AGGREGATOR_NAMESPACE);
        }
        return SourceUtil.createDOMSource(aggregatedDocument);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Aggregator";
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return AGGREGATOR_NAMESPACE;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumes = context.getSUConsumes(exchange.getEndpoint());
            if (consumes != null && consumes.size() == 1) {
                final Document inMessageContent = SourceUtil.createDocument(
                        exchange.getInMessage().getContent());
                // retrieve the document only once to avoid the Streaming issue
                exchange.getInMessage().setContent(new DOMSource(inMessageContent));
                final String correlationID = this.retrieveCorrelationId(inMessageContent, exchange,
                        context);
                if (correlationID != null) {
                    final boolean complete = this.isComplete(inMessageContent, context);
                    if (!this.inMessagesMap.containsKey(correlationID) && !complete) {
                        // TODO implement a timeout on stored exchanges
                        this.inMessagesMap.put(correlationID,
                                new CopyOnWriteArrayList<NormalizedMessage>());
                    }
                    if (complete) {
                        if (this.logger.isLoggable(Level.FINE)) {
                            this.logger.fine(this.getName() + ": sequence is complete");
                        }
                        final Exchange response = this.sendAggregatedMessage(exchange,
                                correlationID, context, consumes.get(0));
                        if (response != null) {
                            final NormalizedMessage receivedMessage = PatternHelper
                                    .processSentExchange(response, context);
                            if (receivedMessage != null) {
                                if (receivedMessage instanceof Fault && !exchange.isInOnlyPattern()) {
                                    exchange.setFault((Fault) receivedMessage);
                                } else if (!exchange.isInOnlyPattern()
                                        && !exchange.isRobustInOnlyPattern()) {
                                    exchange.setOutMessage(receivedMessage);
                                }
                            }
                        } else {
                            exchange.setError(new MessagingException(
                                    "Timeout occurred, no reponse has been received from the target service"));
                        }
                    } else {
                        if (this.logger.isLoggable(Level.FINE)) {
                            this.logger.fine(this.getName() + ": storing exchange '"
                                    + exchange.getExchangeId() + "' for correlation '"
                                    + correlationID + "'");
                        }
                        this.inMessagesMap.get(correlationID).add(exchange.getInMessage());
                        if (this.logger.isLoggable(Level.FINE)) {
                            this.logger.fine(this.getName() + ": sequence size is : "
                                    + this.inMessagesMap.get(correlationID).size());
                        }
                        if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
                            final Document document = PatternHelper
                                    .createResultDocument(AGGREGATOR_NAMESPACE);
                            document.getDocumentElement().setTextContent(
                                    this.getName() + ": the content is buffered by the pattern");
                            exchange.getOutMessage().setContent(
                                    SourceUtil.createStreamSource(document));
                        }
                    }
                } else {
                    exchange.setError(new MessagingException(this.getName()
                            + ": no correlation id found in the message"));
                }
            } else {
                exchange.setError(new MessagingException(this.getName()
                        + ": the Service Unit must define 1 Consumes section"));
            }

        } catch (final MessagingException e) {
            exchange.setError(e);
        } catch (final XPathExpressionException e) {
            exchange.setError(e);
        } catch (PEtALSCDKException e) {
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        exchange.setError(new MessagingException(this.getName()
                + ": the pattern doesn't process the asynchronous exchanges"));
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processMEPValidation(java
     * .net.URI)
     */
    @Override
    public boolean validateMEP(final URI pattern) {
        return true;
    }

    /**
     * @param exchange
     * @return
     * @throws XPathExpressionException
     */
    private boolean isComplete(final Document document, final ExchangeContext context)
            throws XPathExpressionException {
        boolean result = false;
        final String condition = context.getExtensions().get(EIPConstants.TEST);
        result = PatternHelper.evaluateXPathCondition(condition, document.getDocumentElement());
        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine(this.getName() + ": evaluation of condition: \"" + condition
                    + "\", result: \"" + Boolean.toString(result) + "\".");
        }

        return result;
    }

    /**
     * Get the correlation ID from the message with an XPath expression or from
     * the exchange property.
     * 
     * @return
     * @throws JBIException
     * @throws XPathExpressionException
     * @throws XPathExpressionException
     */
    private String retrieveCorrelationId(final Document inMessage, final Exchange exchange,
            final ExchangeContext context) throws XPathExpressionException {
        // retrieve the correlation from the message from an XPATH expression
        String result = null;
        final String expression = context.getExtensions().get(EIPConstants.AGGREGATOR_CORRELATION);
        result = PatternHelper.evaluateXPathExpressionAsString(expression, inMessage);
        if (result != null) {
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.fine(this.getName() + ": evaluation of condition '" + expression
                        + "', result '" + result + "'");
            }
            if (Boolean.FALSE.toString().equals(result) || result.isEmpty()) {
                result = null;
            }
        }

        return result;
    }

    /**
     * @param exchange
     * @param correlationId
     * @param context
     * @param consumes
     * @throws PEtALSCDKException
     * @throws JBIException
     */
    private Exchange sendAggregatedMessage(final Exchange exchange, final String correlationId,
            final ExchangeContext context, final Consumes consumes) throws MessagingException,
            PEtALSCDKException {
        final List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
        if (inMessageList == null) {
            throw new MessagingException(this.getName()
                    + ": no sequence found for the correlation " + correlationId);
        }
        // aggregate the messages into a single one...
        final Source content = buildContent(inMessageList, context);
        // send the aggregated message
        final Exchange response = this.sendExchange(exchange, correlationId, consumes, content,
                context);
        this.inMessagesMap.remove(correlationId);

        return response;
    }

    /**
     * Build the exchange to send to the target service and send it.
     * 
     * @param exchange
     * @param correlationId
     * @param consumes
     * @param content
     * @param context
     * @return the result exchange
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    private Exchange sendExchange(final Exchange exchange, final String correlationId,
            final Consumes consumes, final Source content, final ExchangeContext context)
            throws MessagingException, PEtALSCDKException {
        Exchange newExchange = context.createConsumeExchange(consumes);
        if (newExchange.getOperation() == null) {
            newExchange.setOperation(exchange.getOperation());
        }
        ExchangeUtil.copyExchangeProperties(exchange, newExchange);
        final List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
        for (final NormalizedMessage inMessage : inMessageList) {
            PatternHelper.copy(inMessage, newExchange.getInMessage(), false, true, true, true);
        }
        newExchange.setInMessageContent(content);
        this.logSend(consumes, newExchange);
        if (!context.sendSync(newExchange)) {
            newExchange = null;
        }

        return newExchange;
    }

}
