/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.io.InputStream;
import java.util.List;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * Aggregate all Consumes services defined in the service unit.
 * <p>
 * Return the aggregated results.It accepts InOut only. If a Fault or Error
 * occurs on a call, stop the process and return the Fault or Error if no
 * robustness are used
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class ScatterGather extends AbstractAggregatorPattern {

    private final static String SCATTER_GATHER_NAMESPACE = EIPConstants.NAMESPACE
            + "/scatter-gather";

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Scatter-Gather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractForkerPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return SCATTER_GATHER_NAMESPACE;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean reply = false;
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() != 0) {
                this.sendAsyncExchanges(consumesList, exchange, context);
            } else {
                exchange.setError(new MessagingException(this.getName()
                        + ": the Service Unit must define at least 1 Consumes section(s)"));
                reply = true;
            }
        } catch (final MessagingException e) {
            exchange.setError(e);
            reply = true;
        } catch (PEtALSCDKException e) {
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
            reply = true;
        }
        return reply;
    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Return the list of the Exchanges sent asynchronously. If an error occurs,
     * set a Fault on the original exchange, and return null
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @return the list of the exchanges sent async, or null if a problem
     *         occurred
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    private void sendAsyncExchanges(final List<Consumes> consumesList, final Exchange exchange,
            final ExchangeContext context) throws MessagingException, PEtALSCDKException {
        final boolean faultRobust = this.isFaultRobust(context.getExtensions());
        final boolean exceptionRobust = this.isExceptionRobust(context.getExtensions());
        final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                CommonAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), consumesList.size(),
                faultRobust, exceptionRobust, false);

        final NormalizedMessage inMessage = exchange.getInMessage();
        InputStream inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
        // send all exchanges asynchronously
        for (final Consumes consumedService : consumesList) {
            final Exchange tmpExchange = context.createConsumeExchange(consumedService);
            if (tmpExchange.getOperation() == null) {
                tmpExchange.setOperation(asyncContext.getOriginalExchange().getOperation());
            }
            PatternHelper.copy(inMessage, tmpExchange.getInMessage());
            ExchangeUtil.copyExchangeProperties(exchange, tmpExchange);
            this.logSend(consumedService, tmpExchange);
            context.sendAsync(tmpExchange, asyncContext);

            if (inMessageAsStream != null) {
                // reset the Stream source for next iteration
                inMessage.setContent(new StreamSource(inMessageAsStream));
                inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
            }
        }
    }
}
