/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * Split the source exchange and send the parts to a target service.
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Frederic GARDES - EBM WebSourcing
 */
public class Splitter extends AbstractAggregatorPattern {

    private final static String SPLITTER_NAMESPACE = EIPConstants.NAMESPACE + "/splitter";

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Splitter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean send = false;
        boolean reply = false;
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() == 1) {
                final Consumes consumes = consumesList.get(0);
                List<Source> splittedSources = null;
                if (!this.isAttachmentMode(context.getExtensions())) {
                    // content mode
                    final String splitterPath = context.getExtensions().get(EIPConstants.TEST);
                    if (!StringHelper.isNullOrEmpty(splitterPath)) {
                        if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                            this.logger
                                    .fine(this.getName() + ": splitting XPath : " + splitterPath);
                        }
                        final Document inMessageContent = SourceUtil
                                .createDocument(exchange.getInMessage().getContent());
                        final NodeList nodeList = PatternHelper.evaluateXPathExpressionAsNodeList(
                                splitterPath, inMessageContent);
                        if (nodeList != null && nodeList.getLength() != 0) {
                            if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                                this.logger.fine(this.getName() + ": splitted elements : "
                                        + nodeList.getLength());
                            }
                            splittedSources = new ArrayList<Source>(nodeList.getLength());
                            for (int i = 0; i < nodeList.getLength(); i++) {
                                splittedSources.add(new DOMSource(nodeList.item(i)));
                                send = true;
                            }
                        } else {
                            exchange.setError(new MessagingException(this.getName()
                                    + ": no element to process found"));
                            reply = true;
                        }
                    } else {
                        exchange.setError(new MessagingException(this.getName()
                                + ": the Service Unit must define an element " + EIPConstants.TEST));
                        reply = true;
                    }
                } else {
                    // attachment mode
                    if (exchange.getInMessageAttachments() != null
                            && !exchange.getInMessageAttachments().isEmpty()) {
                        splittedSources = new ArrayList<Source>(exchange.getInMessageAttachments()
                                .size());
                        for (final DataHandler dataHandler : exchange.getInMessageAttachments()) {
                            final String name = dataHandler.getName();
                            final String type = dataHandler.getContentType();
                            if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
                                this.logger.fine(this.getName()
                                        + ": type of the current attachment " + name + ": " + type);
                            }
                            if ("application/xml".equals(type)) {
                                final DataSource content = dataHandler.getDataSource();
                                splittedSources.add(new StreamSource(content.getInputStream()));
                                send = true;
                            } else {
                                this.logger
                                        .info(this.getName()
                                                + ": the attachment "
                                                + name
                                                + " with "
                                                + type
                                                + " as type is ignored. It must be an application/xml type to be process");
                            }
                        }
                    } else {
                        this.logger.info(this.getName() + ": no attachment to process found");
                    }
                }
                // check if we have to send
                if (send) {
                    final boolean faultRobust = this.isFaultRobust(context.getExtensions());
                    final boolean errorRobust = this.isExceptionRobust(context.getExtensions());
                    final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                            CommonAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(),
                            splittedSources.size(), faultRobust, errorRobust, false);
                    for (final Source splittedSource : splittedSources) {
                        final Exchange splittedExchange = context.createConsumeExchange(consumes);
                        if (splittedExchange.getOperation() == null) {
                            splittedExchange.setOperation(exchange.getOperation());
                        }
                        PatternHelper
                                .copy(exchange.getInMessage(), splittedExchange.getInMessage());
                        splittedExchange.getInMessage().setContent(splittedSource);
                        ExchangeUtil.copyExchangeProperties(exchange,
                                splittedExchange);
                        this.logSend(consumes, splittedExchange);
                        context.sendAsync(splittedExchange, asyncContext);
                    }
                }
            } else {
                exchange.setError(new JBIException(this.getName()
                        + ": the Service Unit must define 1 Consumes section"));
                reply = true;
            }
        } catch (final MessagingException e) {
            exchange.setError(e);
            reply = true;
        } catch (final XPathExpressionException e) {
            exchange.setError(e);
            reply = true;
        } catch (final IOException e) {
            exchange.setError(e);
            reply = true;
        } catch (final PEtALSCDKException e) {
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
            reply = true;
        }
        return reply;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractForkerPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return SPLITTER_NAMESPACE;
    }
}
