/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import static org.junit.Assert.fail;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.junit.After;
import org.junit.Before;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.ow2.petals.se.eip.patterns.mock.ExchangeContextMock;
import org.ow2.petals.se.eip.patterns.mock.ExchangeImplMock;
import org.ow2.petals.se.eip.patterns.mock.MessageExchangeMock;
import org.ow2.petals.se.eip.patterns.mock.ServiceEndpointMock;
import org.w3c.dom.Document;

/**
 * @author Frederic Gardes - EBM WebSourcing
 */
public abstract class AbstractTest {

    protected ExchangeContextMock context;

    protected Logger logger;

    /**
     * @param name
     */
    public AbstractTest() {
        
    }

    /**
     * To convert the InputStream to String we use the BufferedReader.readLine()
     * method. We iterate until the BufferedReader return null which means
     * there's no more data to read. Each line will appended to a StringBuilder
     * and returned as String.
     */
    public String convertStreamToString(final InputStream is) {
        final BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        final StringBuilder sb = new StringBuilder();

        String line = null;
        try {
            while ((line = reader.readLine()) != null) {
                sb.append(line + "\n");
            }
        } catch (final IOException e) {
            fail("unable to convert: " + e.getMessage());
        } finally {
            try {
                is.close();
            } catch (final IOException e) {
                fail("unable to convert: " + e.getMessage());
            }
        }
        return sb.toString();
    }

    /**
     * Create a consumer service endpoint
     * 
     * @return a consumer
     */
    public ServiceEndpoint createConsumerServiceEndpoint() {
        final QName[] interfaces = new QName[1];
        QName serviceName = null;
        String endpoint = null;

        interfaces[0] = new QName("http://petals.ow2.org", "interface");
        serviceName = new QName("http://petals.ow2.org", "serviceConsumer");
        endpoint = "8b32ee51-9b7b-4a09-8805-99733c1efc9a";
        final ServiceEndpoint serviceEndpoint = new ServiceEndpointMock(interfaces, serviceName,
                endpoint);
        return serviceEndpoint;
    }

    /**
     * Create a provider service endpoint
     * 
     * @return a provider
     */
    public ServiceEndpoint createProviderServiceEndpoint() {
        final QName[] interfaces = new QName[1];
        QName serviceName = null;
        String endpoint = null;

        interfaces[0] = new QName("http://petals.ow2.org", "interface");
        serviceName = new QName("http://petals.ow2.org", "serviceProvider");
        endpoint = "8b32ee51-9b7b-4a09-8805-99733c1efc9b";
        final ServiceEndpoint serviceEndpoint = new ServiceEndpointMock(interfaces, serviceName,
                endpoint);
        return serviceEndpoint;
    }

    /**
     * Get the list of param from the real test to complete the configuration
     * extensions
     * 
     * @return The Param list
     */
    public abstract List<Param> getParamList();

    /**
     * Create an exchange
     * 
     * @param sourceKind
     *            The source kind: DOM or Stream
     * @param pattern
     *            The pattern
     * @param role
     *            The role
     * @param status
     *            The status
     * @param operation
     *            The operation
     * @param in
     *            The in content
     * @param out
     *            The out content
     * @param fault
     *            The fault content
     * @param exception
     *            The exception
     * @return The exchange created
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    protected ExchangeImplMock createExchange(final SourceKind sourceKind, final URI pattern,
            final Role role, final ExchangeStatus status, final QName operation, final Document in,
            final Document out, final Document fault, final Exception exception)
            throws MessagingException, PEtALSCDKException {
        ExchangeImplMock exchange = null;
        Source source = null;

        exchange = new ExchangeImplMock(new MessageExchangeMock(
                this.createConsumerServiceEndpoint(), pattern, role, status));
        exchange.setEndpoint(this.createProviderServiceEndpoint());
        exchange.setOperation(operation);
        if (in != null) {
            if (sourceKind.isDOM()) {
                source = SourceUtil.createDOMSource(in);
            } else if (sourceKind.isStream()) {
                source = SourceUtil.createStreamSource(in);
            } else {
                throw new PEtALSCDKException("unable to know the source kind to generate");
            }
            exchange.getInMessage().setContent(source);
        }
        if (out != null) {
            if (sourceKind.isDOM()) {
                source = SourceUtil.createDOMSource(out);
            } else if (sourceKind.isStream()) {
                source = SourceUtil.createStreamSource(out);
            } else {
                throw new PEtALSCDKException("unable to know the source kind to generate");
            }
            exchange.getOutMessage().setContent(source);
        }
        if (fault != null) {
            final Fault jbiFault = exchange.createFault();
            if (sourceKind.isDOM()) {
                source = SourceUtil.createDOMSource(fault);
            } else if (sourceKind.isStream()) {
                source = SourceUtil.createStreamSource(fault);
            } else {
                throw new PEtALSCDKException("unable to know the source kind to generate");
            }
            jbiFault.setContent(source);
            exchange.setCheckRole(false);
            exchange.setFault(jbiFault);
            exchange.setCheckRole(true);
        }
        if (exception != null) {
            exchange.setError(exception);
        }
        return exchange;
    }

    /**
     * Create an exchange
     * 
     * @param pattern
     *            The pattern
     * @param role
     *            The role
     * @param status
     *            The status
     * @param in
     *            The in content
     * @param out
     *            The out content
     * @param fault
     *            The fault content
     * @param exception
     *            The exception
     * @return The exchange created
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    protected ExchangeImplMock createExchange(final URI pattern, final Role role,
            final ExchangeStatus status, final Document in, final Document out,
            final Document fault, final Exception exception) throws MessagingException,
            PEtALSCDKException {
        final SourceKind sourceKind = new SourceKind();
        sourceKind.setStreamKind();
        return this.createExchange(sourceKind, pattern, role, status, null, in, out, fault,
                exception);
    }

    protected Exchange processAsyncDOMTest(final URI pattern, final Role role,
            final ExchangeStatus status, final Document in, final Document out,
            final Document fault, final Exception exception, final AbstractPattern abstractPattern,
            final CommonAsyncContext eipAsyncContext) throws MessagingException, PEtALSCDKException {
        final SourceKind sourceKind = new SourceKind();
        sourceKind.setDOMKind();
        final ExchangeImplMock exchange = this.createExchange(sourceKind, pattern, role, status,
                null, in, out, fault, exception);
        abstractPattern.logger = this.logger;
        abstractPattern.processAsync(exchange, this.context, eipAsyncContext);
        return exchange;
    }

    protected Exchange processAsyncStreamTest(final URI pattern, final Role role,
            final ExchangeStatus status, final Document in, final Document out,
            final Document fault, final Exception exception, final AbstractPattern abstractPattern,
            final CommonAsyncContext eipAsyncContext) throws MessagingException, PEtALSCDKException {
        final SourceKind sourceKind = new SourceKind();
        sourceKind.setStreamKind();
        final ExchangeImplMock exchange = this.createExchange(sourceKind, pattern, role, status,
                null, in, out, fault, exception);
        abstractPattern.logger = this.logger;
        abstractPattern.processAsync(exchange, this.context, eipAsyncContext);
        return exchange;
    }

    protected Exchange processDOMTest(final URI pattern, final Role role,
            final ExchangeStatus status, final QName operation, final Document in,
            final Document out, final Document fault, final Exception exception,
            final AbstractPattern abstractPattern) throws MessagingException, PEtALSCDKException {
        final SourceKind sourceKind = new SourceKind();
        sourceKind.setDOMKind();
        final ExchangeImplMock exchange = this.createExchange(sourceKind, pattern, role, status,
                operation, in, out, fault, exception);
        abstractPattern.logger = this.logger;
        abstractPattern.process(exchange, this.context);
        return exchange;
    }

    protected Exchange processStreamTest(final URI pattern, final Role role,
            final ExchangeStatus status, final QName operation, final Document in,
            final Document out, final Document fault, final Exception exception,
            final AbstractPattern abstractPattern) throws MessagingException, PEtALSCDKException {
        final SourceKind sourceKind = new SourceKind();
        sourceKind.setStreamKind();
        final ExchangeImplMock exchange = this.createExchange(sourceKind, pattern, role, status,
                operation, in, out, fault, exception);
        abstractPattern.logger = this.logger;
        abstractPattern.process(exchange, this.context);
        return exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see junit.framework.TestCase#setUp()
     */
    @Before
    public void setUp() throws Exception {
        this.context = new ExchangeContextMock(new ConfigurationExtensions(this.getParamList()));
        this.logger = Logger.getLogger(this.getClass().getSimpleName());
    }

    /*
     * (non-Javadoc)
     * 
     * @see junit.framework.TestCase#tearDown()
     */
    @After
    public void tearDown() throws Exception {
        this.context = null;
    }
}
